# Python Notebook: Bike Store Sample Database - SQL Concepts in Python

# This notebook demonstrates how to perform SQL queries and concepts
# using Python, primarily with the pandas library, on the Bike Store Sample Database.

# Prerequisites:
# 1. Ensure you have the 'bike_store.sqlite' database file in the same directory as this notebook,
#    or provide the full path to the file.
# 2. Install pandas and sqlite3 (sqlite3 is usually built-in with Python).
#    pip install pandas

import pandas as pd
import sqlite3
import numpy as np

# --- Database Connection Setup ---
# Replace 'bike_store.sqlite' with the actual path to your database file if it's not in the same directory.
DB_FILE = 'bike_store.sqlite'

try:
    conn = sqlite3.connect(DB_FILE)
    print(f"Successfully connected to {DB_FILE}")
except sqlite3.Error as e:
    print(f"Error connecting to database: {e}")
    # Exit or handle the error appropriately if connection fails
    exit()

# Helper function to run SQL queries and display results as a DataFrame
def run_sql_query(query, description=""):
    print(f"\n--- {description} ---")
    print(f"SQL Query:\n{query}")
    try:
        df = pd.read_sql_query(query, conn)
        print("Result:")
        print(df)
    except pd.io.sql.DatabaseError as e:
        print(f"Error executing query: {e}")
    print("-" * 50)

# --- 1. Basic SQL Queries in Python ---

print("\n### 1. Basic SQL Queries ###")

# 1.1. SELECT, FROM, WHERE
run_sql_query(
    "SELECT * FROM customers LIMIT 5;",
    "Select all columns from the 'customers' table (first 5 rows)"
)

run_sql_query(
    "SELECT first_name, last_name, email FROM customers LIMIT 5;",
    "Select specific columns (first_name, last_name, email) from 'customers' (first 5 rows)"
)

run_sql_query(
    "SELECT first_name, last_name, city FROM customers WHERE city = 'New York';",
    "Select customers from 'New York' city"
)

run_sql_query(
    "SELECT product_name, list_price FROM products WHERE list_price > 1500 LIMIT 5;",
    "Select products with a list_price greater than 1500 (first 5 rows)"
)

# 1.2. ORDER BY, LIMIT
run_sql_query(
    "SELECT product_name, list_price FROM products ORDER BY list_price ASC LIMIT 5;",
    "Select products ordered by list_price in ascending order (first 5)"
)

run_sql_query(
    "SELECT product_name, list_price FROM products ORDER BY list_price DESC LIMIT 5;",
    "Select the 5 most expensive products"
)

run_sql_query(
    "SELECT first_name, last_name, state FROM customers WHERE state = 'CA' ORDER BY last_name ASC, first_name ASC LIMIT 5;",
    "Select customers from 'California' ordered by last_name then first_name (first 5)"
)

# 1.3. DISTINCT
run_sql_query(
    "SELECT DISTINCT city FROM customers LIMIT 10;",
    "Get all unique cities where customers are located (first 10)"
)

run_sql_query(
    "SELECT DISTINCT state FROM stores;",
    "Get all unique states where stores are located"
)

# 1.4. Aggregate Functions (COUNT, SUM, AVG, MIN, MAX)
run_sql_query(
    "SELECT COUNT(*) AS total_customers FROM customers;",
    "Count the total number of customers"
)

run_sql_query(
    "SELECT AVG(list_price) AS average_product_price FROM products;",
    "Calculate the average list price of all products"
)

run_sql_query(
    "SELECT MIN(list_price) AS cheapest_product, MAX(list_price) AS most_expensive_product FROM products;",
    "Find the minimum and maximum list price of products"
)

run_sql_query(
    "SELECT SUM(quantity) AS total_items_sold FROM order_items;",
    "Calculate the total quantity of items sold in all orders"
)

# 1.5. GROUP BY, HAVING
run_sql_query(
    "SELECT city, COUNT(customer_id) AS customer_count FROM customers GROUP BY city ORDER BY customer_count DESC LIMIT 5;",
    "Count the number of customers in each city (top 5)"
)

run_sql_query(
    """
    SELECT product_id, SUM(quantity * list_price) AS total_revenue_per_product
    FROM order_items
    GROUP BY product_id
    ORDER BY total_revenue_per_product DESC
    LIMIT 5;
    """,
    "Calculate the total sales (quantity * list_price) for each product (top 5)"
)

run_sql_query(
    "SELECT category_id, COUNT(product_id) AS product_count FROM products GROUP BY category_id HAVING COUNT(product_id) > 10;",
    "Find categories with more than 10 products"
)

run_sql_query(
    """
    SELECT store_id, SUM(quantity * list_price * (1 - discount)) AS total_store_revenue
    FROM order_items oi
    JOIN orders o ON oi.order_id = o.order_id
    GROUP BY store_id
    HAVING total_store_revenue > 10000
    ORDER BY total_store_revenue DESC;
    """,
    "Calculate the total revenue for each store, only for stores with total revenue > 10000"
)

# --- 2. Intermediate SQL Techniques in Python ---

print("\n### 2. Intermediate SQL Techniques ###")

# 2.1. JOIN Types
run_sql_query(
    """
    SELECT c.first_name, c.last_name, o.order_id, o.order_date
    FROM customers AS c
    INNER JOIN orders AS o ON c.customer_id = o.customer_id
    ORDER BY o.order_date DESC
    LIMIT 10;
    """,
    "Get customer name and their order IDs (INNER JOIN)"
)

run_sql_query(
    """
    SELECT p.product_name, b.brand_name
    FROM products AS p
    INNER JOIN brands AS b ON p.brand_id = b.brand_id
    LIMIT 10;
    """,
    "Get product name and its brand name (INNER JOIN)"
)

run_sql_query(
    """
    SELECT s.store_name, st.first_name, st.last_name
    FROM stores AS s
    LEFT JOIN staffs AS st ON s.store_id = st.store_id
    ORDER BY s.store_name, st.first_name
    LIMIT 10;
    """,
    "Get all stores and the staff members associated with them (LEFT JOIN)"
)

run_sql_query(
    """
    SELECT s.first_name AS staff_first_name,
           s.last_name AS staff_last_name,
           m.first_name AS manager_first_name,
           m.last_name AS manager_last_name
    FROM staffs AS s
    LEFT JOIN staffs AS m ON s.manager_id = m.staff_id
    ORDER BY staff_last_name
    LIMIT 10;
    """,
    "Example of a self-join: Find staff members and their managers"
)

# 2.2. UNION, UNION ALL
run_sql_query(
    """
    SELECT email FROM customers
    UNION
    SELECT email FROM staffs
    LIMIT 20;
    """,
    "Get a combined list of customer emails and staff emails (UNION - removes duplicates)"
)

run_sql_query(
    """
    SELECT email FROM customers
    UNION ALL
    SELECT email FROM staffs
    LIMIT 20;
    """,
    "Get a combined list of customer emails and staff emails (UNION ALL - keeps duplicates)"
)

# 2.3. Subqueries
run_sql_query(
    """
    SELECT product_name, list_price
    FROM products
    WHERE list_price > (SELECT AVG(list_price) FROM products)
    ORDER BY list_price DESC
    LIMIT 5;
    """,
    "Scalar Subquery: Find products whose list_price is higher than the average product price (top 5)"
)

run_sql_query(
    """
    SELECT first_name, last_name
    FROM customers
    WHERE customer_id IN (SELECT DISTINCT customer_id FROM orders)
    LIMIT 10;
    """,
    "Multi-row Subquery with IN: Find customers who have placed at least one order (first 10)"
)

run_sql_query(
    """
    SELECT first_name, last_name
    FROM customers
    WHERE customer_id NOT IN (SELECT DISTINCT customer_id FROM orders)
    LIMIT 10;
    """,
    "Multi-row Subquery with NOT IN: Find customers who have NOT placed any orders (first 10)"
)

run_sql_query(
    """
    SELECT s.store_name, store_revenue.total_revenue
    FROM stores AS s
    JOIN (
        SELECT store_id, SUM(quantity * list_price * (1 - discount)) AS total_revenue
        FROM order_items oi
        JOIN orders o ON oi.order_id = o.order_id
        GROUP BY store_id
    ) AS store_revenue ON s.store_id = store_revenue.store_id
    ORDER BY store_revenue.total_revenue DESC;
    """,
    "Subquery in FROM clause (Derived Table): Find the total revenue per store"
)

# 2.4. CASE Statements (using pandas.apply or numpy.select for conditional logic)
# First, load the products table to demonstrate
products_df = pd.read_sql_query("SELECT product_name, list_price FROM products;", conn)

# Categorize products based on their list_price using numpy.select
conditions = [
    products_df['list_price'] < 500,
    (products_df['list_price'] >= 500) & (products_df['list_price'] < 1500),
    (products_df['list_price'] >= 1500) & (products_df['list_price'] < 3000),
    products_df['list_price'] >= 3000
]
choices = ['Budget', 'Mid-Range', 'Premium', 'Luxury']
products_df['price_category'] = np.select(conditions, choices, default='Unknown')

print("\n--- Categorize products based on their list_price (Python/Pandas) ---")
print(products_df.head(10))
print("-" * 50)

# Display order status in a more readable format
orders_df = pd.read_sql_query("SELECT order_id, order_date, order_status FROM orders LIMIT 10;", conn)
status_map = {1: 'Pending', 2: 'Processing', 3: 'Rejected', 4: 'Completed'}
orders_df['readable_order_status'] = orders_df['order_status'].map(status_map).fillna('Unknown')

print("\n--- Display order status in a more readable format (Python/Pandas) ---")
print(orders_df[['order_id', 'order_date', 'readable_order_status']])
print("-" * 50)


# 2.5. LIKE Operator (using pandas string methods)
customers_df = pd.read_sql_query("SELECT first_name, last_name FROM customers;", conn)

print("\n--- Find customers whose first name starts with 'J' (Python/Pandas) ---")
print(customers_df[customers_df['first_name'].str.startswith('J')].head(10))
print("-" * 50)

print("\n--- Find products with 'mountain' in their name (Python/Pandas) ---")
products_df_all = pd.read_sql_query("SELECT product_name FROM products;", conn)
print(products_df_all[products_df_all['product_name'].str.contains('mountain', case=False)].head(10))
print("-" * 50)

print("\n--- Find customers whose last name has 'a' as the second letter (Python/Pandas) ---")
print(customers_df[customers_df['last_name'].str.match(r'^.a')].head(10)) # Regex for second letter 'a'
print("-" * 50)


# 2.6. IN, NOT IN, BETWEEN
run_sql_query(
    """
    SELECT p.product_name, b.brand_name
    FROM products p
    JOIN brands b ON p.brand_id = b.brand_id
    WHERE b.brand_name IN ('Trek', 'Electra')
    LIMIT 10;
    """,
    "Find products from specific brands ('Trek', 'Electra')"
)

run_sql_query(
    """
    SELECT order_id, order_date, order_status
    FROM orders
    WHERE order_status NOT IN (1, 3)
    LIMIT 10;
    """,
    "Find orders that are not in 'Pending' or 'Rejected' status"
)

run_sql_query(
    """
    SELECT product_name, list_price
    FROM products
    WHERE list_price BETWEEN 1000 AND 2000
    LIMIT 10;
    """,
    "Find products with list_price between 1000 and 2000 (inclusive)"
)

run_sql_query(
    """
    SELECT order_id, order_date
    FROM orders
    WHERE order_date BETWEEN '2016-01-01' AND '2016-12-31'
    ORDER BY order_date
    LIMIT 10;
    """,
    "Find orders placed between two specific dates"
)

# --- 3. Advanced SQL Techniques in Python ---

print("\n### 3. Advanced SQL Techniques ###")

# 3.1. Window Functions (using pandas.groupby().transform() or .rank(), .shift())
products_all_df = pd.read_sql_query("SELECT product_name, category_id, model_year, list_price FROM products ORDER BY category_id, model_year, product_name;", conn)

# ROW_NUMBER(), RANK(), DENSE_RANK()
products_all_df['row_num_in_category'] = products_all_df.groupby('category_id')['list_price'].rank(method='first', ascending=False)
products_all_df['rank_in_category'] = products_all_df.groupby('category_id')['list_price'].rank(method='min', ascending=False)
products_all_df['dense_rank_in_category'] = products_all_df.groupby('category_id')['list_price'].rank(method='dense', ascending=False)

print("\n--- ROW_NUMBER(), RANK(), DENSE_RANK() (Python/Pandas) ---")
print(products_all_df[['product_name', 'category_id', 'list_price', 'row_num_in_category', 'rank_in_category', 'dense_rank_in_category']].head(20))
print("-" * 50)

# LAG() and LEAD()
products_all_df['previous_year_price'] = products_all_df.groupby('category_id')['list_price'].shift(1)
products_all_df['price_difference'] = products_all_df['list_price'] - products_all_df['previous_year_price'].fillna(0)

print("\n--- LAG() and LEAD() (Python/Pandas) ---")
print(products_all_df[['product_name', 'category_id', 'model_year', 'list_price', 'previous_year_price', 'price_difference']].head(20))
print("-" * 50)

# Aggregate Window Functions: Running total of quantity sold per product
order_items_df = pd.read_sql_query("SELECT order_id, product_id, quantity FROM order_items ORDER BY product_id, order_id;", conn)
order_items_df['running_total_quantity'] = order_items_df.groupby('product_id')['quantity'].cumsum()

print("\n--- Aggregate Window Functions: Running total of quantity sold per product (Python/Pandas) ---")
print(order_items_df.head(20))
print("-" * 50)

# NTILE() (simulated using qcut)
customers_orders_df = pd.read_sql_query(
    "SELECT c.customer_id, c.first_name, c.last_name, COUNT(o.order_id) AS total_orders FROM customers c JOIN orders o ON c.customer_id = o.customer_id GROUP BY c.customer_id, c.first_name, c.last_name ORDER BY total_orders DESC;",
    conn
)
# qcut divides data into n quantiles. labels=False makes it return integer labels (0 to n-1)
# Add 1 to make it 1 to n, similar to NTILE
customers_orders_df['order_quartile'] = pd.qcut(customers_orders_df['total_orders'], 4, labels=False, duplicates='drop') + 1

print("\n--- NTILE() (simulated with Pandas qcut) ---")
print(customers_orders_df[['first_name', 'last_name', 'total_orders', 'order_quartile']].head(20))
print("-" * 50)


# 3.2. Common Table Expressions (CTEs) - Can be done by chaining pandas operations or using SQL directly
# Example: Find the top 3 products by total revenue per category (using direct SQL with CTE)
run_sql_query(
    """
    WITH ProductRevenue AS (
        SELECT
            p.product_id,
            p.product_name,
            c.category_name,
            SUM(oi.quantity * oi.list_price * (1 - oi.discount)) AS total_product_revenue
        FROM order_items oi
        JOIN products p ON oi.product_id = p.product_id
        JOIN categories c ON p.category_id = c.category_id
        GROUP BY p.product_id, p.product_name, c.category_name
    ),
    RankedProductRevenue AS (
        SELECT
            product_name,
            category_name,
            total_product_revenue,
            ROW_NUMBER() OVER (PARTITION BY category_name ORDER BY total_product_revenue DESC) AS rn
        FROM ProductRevenue
    )
    SELECT
        product_name,
        category_name,
        total_product_revenue
    FROM RankedProductRevenue
    WHERE rn <= 3
    ORDER BY category_name, total_product_revenue DESC;
    """,
    "CTE Example: Find the top 3 products by total revenue per category (using SQL directly)"
)

# Example: Calculate monthly sales for each store (using pandas operations)
orders_items_stores_df = pd.read_sql_query(
    """
    SELECT o.order_date, o.store_id, oi.quantity, oi.list_price, oi.discount, s.store_name
    FROM orders o
    JOIN order_items oi ON o.order_id = oi.order_id
    JOIN stores s ON o.store_id = s.store_id;
    """,
    conn
)
orders_items_stores_df['order_date'] = pd.to_datetime(orders_items_stores_df['order_date'])
orders_items_stores_df['sales_month'] = orders_items_stores_df['order_date'].dt.strftime('%Y-%m')
orders_items_stores_df['revenue'] = orders_items_stores_df['quantity'] * orders_items_stores_df['list_price'] * (1 - orders_items_stores_df['discount'])

monthly_sales_df = orders_items_stores_df.groupby(['sales_month', 'store_id', 'store_name'])['revenue'].sum().reset_index()
monthly_sales_df.rename(columns={'revenue': 'monthly_revenue'}, inplace=True)
monthly_sales_df.sort_values(by=['sales_month', 'store_name'], inplace=True)

print("\n--- CTE Example: Calculate monthly sales for each store (Python/Pandas) ---")
print(monthly_sales_df.head(10))
print("-" * 50)


# 3.3. Conditional Aggregation (Simulating PIVOT)
# Load necessary data
orders_stores_df = pd.read_sql_query(
    """
    SELECT s.store_name, o.order_id, o.order_status
    FROM stores s
    LEFT JOIN orders o ON s.store_id = o.store_id;
    """,
    conn
)

# Count completed vs. pending orders per store using pivot_table
# Note: order_status 1=Pending, 4=Completed
pivot_df = orders_stores_df.pivot_table(
    index='store_name',
    columns='order_status',
    values='order_id',
    aggfunc='count',
    fill_value=0
)

# Rename columns for clarity
pivot_df.rename(columns={1: 'pending_orders', 4: 'completed_orders'}, inplace=True)

# Add total_orders
pivot_df['total_orders'] = pivot_df[['pending_orders', 'completed_orders']].sum(axis=1)

print("\n--- Count completed vs. pending orders per store (Python/Pandas Pivot) ---")
print(pivot_df)
print("-" * 50)

# --- Close Database Connection ---
conn.close()
print(f"\nDisconnected from {DB_FILE}")

# --- Conclusion ---
# This notebook has demonstrated how to translate various SQL concepts, from basic queries
# to advanced techniques like window functions and CTEs, into Python using the pandas library.
# This approach allows for powerful data manipulation and analysis within a Python environment,
# leveraging both SQL query capabilities and pandas' rich DataFrame functionalities.
#
# Next Steps:
# - Experiment with more complex data transformations in pandas.
# - Integrate data visualization libraries (e.g., matplotlib, seaborn, plotly) to visualize results.
# - Explore more advanced SQL concepts like recursive CTEs or database-specific functions.
